"""
Sphinx/docutils extension to create links to pyDoctor documentation using a
RestructuredText interpreted text role that looks like this::

    :api:`python_object_to_link_to <label>`

for example::

    :api:`twisted.internet.defer.Deferred <Deferred>`

"""

from types import MappingProxyType


apilinks_base_url = "https://docs.twisted.org/en/stable/api/"

emptyMapping = MappingProxyType({})


def make_api_link(
    name,
    rawtext,
    text,
    lineno,
    inliner,
    options=emptyMapping,
    content=(),
):
    from docutils import nodes, utils

    # quick, dirty, and ugly...
    if "<" in text and ">" in text:
        full_name, label = text.split("<")
        full_name = full_name.strip()
        label = label.strip(">").strip()
    else:
        full_name = text

    # not really sufficient, but just testing...
    # ...hmmm, maybe this is good enough after all
    ref = "".join((apilinks_base_url, full_name, ".html"))

    node = nodes.reference(
        rawtext, utils.unescape(label), refuri=ref, **options
    )

    nodes = [node]
    sys_msgs = []
    return nodes, sys_msgs


# setup function to register the extension


def setup(app):
    app.add_config_value(
        "apilinks_base_url",
        "https://docs.twisted.org/en/stable/api/",
        "env",
    )
    app.add_role("api", make_api_link)



from hashlib import sha256
from random import Random
from struct import unpack

from twisted.web.template import slot, tags

from klein import Klein, Plating


# Cobble together a deterministic random function using a string as a seed.
def random_from_string(string):
    return Random(unpack("!I", sha256(string.encode("utf-8")).digest()[:4])[0])


app = Klein()

myStyle = Plating(
    tags=tags.html(
        tags.head(tags.title(slot("pageTitle"))),
        tags.body(
            tags.h1(slot("pageTitle"), Class="titleHeading"),
            tags.div(slot(Plating.CONTENT)),
        ),
    ),
    defaults={"pageTitle": "Places & Foods"},
)


@myStyle.routed(
    app.route("/"),
    tags.div(
        tags.h2("Sample Places:"),
        tags.ul(
            [
                tags.li(tags.a(href=["/places/", place])(place))
                for place in ["new york", "san francisco", "shanghai"]
            ]
        ),
        tags.h2("Sample Foods:"),
        tags.ul(
            [
                tags.li(tags.a(href=["/foods/", food])(food))
                for food in ["hamburgers", "cheeseburgers", "hot dogs"]
            ]
        ),
    ),
)
def root(request):
    return {}


@myStyle.routed(
    app.route("/foods/<food>"),
    tags.table(border="2", style="color: blue")(
        tags.tr(tags.td("Name:"), tags.td(slot("name"))),
        tags.tr(tags.td("Deliciousness:"), tags.td(slot("rating"), " stars")),
        tags.tr(tags.td("Carbs:"), tags.td(slot("carbohydrates"))),
    ),
)
def one_food(request, food):
    random = random_from_string(food)
    return {
        "name": food,
        "pageTitle": f"Food: {food}",
        "rating": random.randint(1, 5),
        "carbohydrates": random.randint(0, 100),
    }


@myStyle.routed(
    app.route("/places/<place>"),
    tags.div(style="color: green")(
        tags.h1("Place: ", slot("name")),
        tags.div(slot("latitude"), "° ", slot("longitude"), "°"),
        tags.div(
            tags.h2("Foods Found Here:"),
            tags.ul(
                tags.li(render="foods:list")(
                    tags.a(href=["/foods/", slot("item")])(slot("item"))
                )
            ),
        ),
    ),
)
def one_place(request, place):
    random = random_from_string(place)
    possible_foods = [
        "hamburgers",
        "cheeseburgers",
        "hot dogs",
        "pizza",
        "叉烧",
        "皮蛋",
        "foie gras",
    ]
    random.shuffle(possible_foods)
    return {
        "name": place,
        "pageTitle": f"Place: {place}",
        "latitude": random.uniform(-90, 90),
        "longitude": random.uniform(-180, 180),
        "foods": possible_foods[:3],
    }


app.run("localhost", 8080)



from twisted.web.template import slot, tags

from klein import Field, Form, Klein, Plating, Requirer, SessionProcurer
from klein.interfaces import ISession
from klein.storage.memory import MemorySessionStore


app = Klein()

sessions = MemorySessionStore()

requirer = Requirer()


@requirer.prerequisite([ISession])
def procurer(request):
    return SessionProcurer(sessions).procureSession(request)


style = Plating(
    tags=tags.html(
        tags.head(tags.title("yay")), tags.body(tags.div(slot(Plating.CONTENT)))
    )
)


@requirer.require(
    style.routed(
        app.route("/", methods=["POST"]),
        tags.h1("u did it: ", slot("an-form-arg")),
    ),
    foo=Field.number(minimum=3, maximum=10),
    bar=Field.text(),
)
def postHandler(foo, bar):
    return {"an-form-arg": foo}


@requirer.require(
    style.routed(app.route("/", methods=["GET"]), tags.div(slot("anForm"))),
    theForm=Form.rendererFor(postHandler, action="/?post=yes"),
)
def formRenderer(theForm):
    return {"anForm": theForm}


@requirer.require(
    style.routed(
        Form.onValidationFailureFor(postHandler),
        [tags.h1("invalid form"), tags.div(slot("the-invalid-form"))],
    ),
    renderer=Form.rendererFor(postHandler, action="/?post=yes"),
)
def validationFailed(values, renderer):
    renderer.prevalidationValues = values.prevalidationValues
    renderer.validationErrors = values.validationErrors
    return {"the-invalid-form": renderer}


app.run("localhost", 8080)



from __future__ import annotations

import attrs
from zope.interface import Interface


@attrs.define(repr=False)
class _ProvidesValidator:
    interface: type[Interface] = attrs.field()

    def __call__(
        self, inst: object, attr: attrs.Attribute, value: object
    ) -> None:
        """
        We use a callable class to be able to change the ``__repr__``.
        """
        if not self.interface.providedBy(value):
            msg = (
                f"'{attr.name}' must provide {self.interface!r} "
                f"which {value!r} doesn't."
            )
            raise TypeError(
                msg,
                attr,
                self.interface,
                value,
            )

    def __repr__(self) -> str:
        return f"<provides validator for interface {self.interface!r}>"


def provides(interface: type[Interface]) -> _ProvidesValidator:
    """
    A validator that raises a `TypeError` if the initializer is called
    with an object that does not provide the requested *interface* (checks are
    performed using ``interface.providedBy(value)`` (see `zope.interface
    <https://zopeinterface.readthedocs.io/en/latest/>`_).

    :param interface: The interface to check for.
    :type interface: ``zope.interface.Interface``

    :raises TypeError: With a human readable error message, the attribute
        (of type `attrs.Attribute`), the expected interface, and the
        value it got.
    """
    return _ProvidesValidator(interface)



from __future__ import annotations

import attrs
from zope.interface import Interface, implementer

from twisted.trial.unittest import SynchronousTestCase

from .._attrs_zope import provides


class IWhatever(Interface):
    ...


@implementer(IWhatever)
class YesWhatever:
    ...


class NoWhatever:
    ...


@attrs.define()
class WhateverContainer:
    whatever: object = attrs.field(validator=provides(IWhatever))


class ProvidesTestCase(SynchronousTestCase):
    def test_yes(self) -> None:
        WhateverContainer(YesWhatever())

    def test_no(self) -> None:
        with self.assertRaises(TypeError):
            WhateverContainer(NoWhatever())

    def test_repr(self) -> None:
        self.assertIn("provides validator for", repr(provides(IWhatever)))
        self.assertIn("IWhatever", repr(provides(IWhatever)))
